// Compile in MinGW with the following command:
// g++ -o template.dll template.cpp -W -Wall -shared -Wl,--enable-auto-import -Wl,--enable-runtime-pseudo-reloc -s -lalleg
// and in Linux with
// g++ -o template.dll template.cpp -W -Wall -shared -s -fPIC `allegro-config --libs`
// ...or something like that...

#include "IGame.h"
#include <assert.h>
#include <list>

// # of tiles
#define TILENUM 5

// size of map
#define MAPW 32
#define MAPH 26

// size of tiles
#define TILEW 16
#define TILEH 16

// start location of pacman, in map coordinates
#define PACMAN_STARTX 16
#define PACMAN_STARTY 19

// start location of ghosts, in map coordinates
#define GHOST_STARTX 16
#define GHOST_STARTY 12

// interval between logic ticks, in msec
#define INTERVAL 25

// macro
#define BOUND(x,y,z) ((x) < (y) ? (y) : ((x) > (z) ? (z) : (x)))

// tilemap offset from top left of screen
#define XOFST 64
#define YOFST 0

// duration of power pill effect
#define POWERTIME 300
// duration of level intro
#define INTROTIME 100


static volatile int timer = 0;
static void timer_f() {
	++timer;
}
END_OF_STATIC_FUNCTION(timer_f);

using namespace std;

class PcManObject {
	
	public:
		PcManObject() : oldx (0), oldy (0), dirty (false), dx(1), dy(0), x(0), y(0) {}

		// x    : bits 0..7,  mask 255
		// y    : bits 8..15, mask 65280 
		// mx  : bits 3..7,  mask 248
		// my  : bits 11..15, mask 63488	

		int getx() { return x; }
		int gety() { return y; }
		int getmx() { return x / TILEW; }
		int getmy() { return y / TILEH; }
	
		void setx (int newx) 
		{ 
			if (newx != x) 
			{ 
				if (!dirty) { oldx = x; } 
				x = newx; dirty = true; 
			} 
		}
		void sety (int newy) 
		{ 
			if (newy != y) 
			{ 
				if (!dirty) { oldy = y; }
				y = newy; dirty = true; 
			} 
		}
		void move (int _dx, int _dy) 
		{ 
			if (_dx != 0 || _dy != 0) 
			{ 
				if (!dirty) { oldx = x; oldy = y; }
				x += _dx; 
				y += _dy; 
				dirty = true; 
			} 
		}
		
		int oldx; int oldy;
		bool dirty;
		int dx, dy;
		enum Type {PACMAN, GHOST, FRUIT, SCORE};
		Type type;
		int sprite;
		int scared; // for ghosts only
	private:
		int x, y; // position
};

class PcMan : public IGame {
	private:
		static const int mapDefault [MAPW * MAPH];
			
		int introtime; // if introtime - time > 0, we're displaying intro.
		int riptime; // if riptime - time > 0, short break before restoring pacman
		bool paused;
		bool screenDirty;
		int m[MAPW * MAPH];
		int time;
		
		list<PcManObject> lobj;
		PcManObject *player;
		int score;
		int left;
		int lives;
		int level;
		int multiplier;
		
		FONT *qfont;
		BITMAP *sprites;
		
	public:
		PcMan() : IGame(), screenDirty (true), sprites (NULL) {
		}

		~PcMan() {
		}

		char *GetName() {
			return "Pacman 729";
		}

		char *GetAuthor() {
			return "Amarillion";
		}

		char *GetDescription() {
			return "Eat all the pills while avoiding the ghosts. \
If you eat a power pill, the ghosts run away from you instead.";
		}

		char *GetVersion() {
			return "0.95";
		}

		char *GetIconPath() {
			return "pacman.bmp";
		}

		void InitObjects()
		{
			lobj.clear();
			// start locations of objects
			{
				PcManObject o;
				o.setx(PACMAN_STARTX * TILEW); 
				o.sety(PACMAN_STARTY * TILEH);
				o.dx = 1; o.dy = 0;
				o.type = PcManObject::PACMAN;
				o.sprite = 0;
				o.scared = -1;
				lobj.push_back (o);
			}
			player = &(*(lobj.begin()));
			int i;
			for (i = 0; i < 4; ++i)
			{
				PcManObject o;
				o.setx(GHOST_STARTX * TILEW); 
				o.sety(GHOST_STARTY * TILEH);
				o.dx = 1; o.dy = 0;
				o.type = PcManObject::GHOST;
				o.sprite = i + 1;
				o.scared = -1;
				lobj.push_back (o);
			}
		}
		
		void InitLevel() 
		{
			introtime = time + INTROTIME; // display intro 100 steps
			riptime = -1;
			multiplier = 100;
			screenDirty = true;
			
			InitObjects();
			
			// reset map, symmetrically
			left = 0;
			int i;
			for (i = 0; i < MAPH * MAPW / 2; ++i)
			{
				m[i] = mapDefault[i];
				m[(25-(i / 32))*32 + (i & 31)] = mapDefault[i];
				
				if (m[i] == 0) left += 2; // count number of pills to eat
			}	
		}
		
		bool Init() {
			if (!IGame::Init()) {
				return false;
			}
			
			LOCK_VARIABLE(timer);
			LOCK_FUNCTION(timer_f);
			install_int(timer_f, INTERVAL);

			// load font
			qfont = load_font("games/8x8.pcx", 0, 0);
			if (!qfont) {
				qfont = font;
			}

			// init graphics
			const char *tile[TILENUM] = {".","*","o"," ","|"};
			const char *a[] = {"C","M","M","M","M"};
			const char *a2[] = {"C","W","W","W","W"};
			int c[TILENUM]; // colors of tiles
			int e[5]; // colors of objects

			// initialize tile colors
			c[0] = makecol (255, 255, 255); // '.'
			c[1] = makecol (128, 128, 128); // '*'
			c[2] = makecol (255, 255, 0);   // 'o'
			c[3] = makecol (0, 0, 0);       // ' '
			c[4] = makecol (255, 255, 255); // '|'
			// initialize object colors
			e[0] = makecol (255, 255, 0); // pacman
			e[1] = makecol (0, 0, 255);   // ghost 1
			e[2] = makecol (0, 255, 0);   // ghost 2
			e[3] = makecol (255, 0, 0);   // ghost 3
			e[4] = makecol (0, 255, 255); // ghost 4
			
			sprites = create_bitmap (160, 160);
			if (!sprites) return false;
			clear_bitmap (sprites);
			BITMAP *temp = create_bitmap (80, 80);
			int i;
			for (i = 0; i < 5; ++i)
			{
				textout_ex (temp, qfont, 
					tile[i], 
					i * 8, 0, 
					c[i], makecol(0,0,0));	
				textout_ex (temp, qfont, 
					a[i], 
					i * 8, 8, 
					e[i], makecol(0,0,0));	
				textout_ex (temp, qfont, 
					a2[i],
					i * 8, 16, 
					e[i], makecol(0,0,0));	
			}
			stretch_blit (temp, sprites, 0, 0, temp->w, temp->h, 
				0, 0, sprites->w, sprites->h);
			destroy_bitmap (temp);
			
			// initialize other variables
			time = 0;
			score = 0;
			lives = 4;
			paused = false;
						
			level = 1;
			InitLevel();
			
			return true;
		}

		void Deinit() {
			if (qfont != font) {
				destroy_font(qfont);
				qfont = font;
			}
			
			// no more Allegro calls after this line!
			IGame::Deinit();
		}

		void DrawBackground (BITMAP *canvas, int x1, int y1, int x2, int y2) 
		{			
			int xco, yco;
			rectfill (canvas, x1, y1, x2, y2, makecol (0,0,0));
			for (xco = BOUND ((x1 - XOFST) / TILEW, 0, MAPW-1); 
				 xco <= BOUND ((x2 - XOFST) / TILEW, 0, MAPW-1); ++xco)
			{
				for (yco = BOUND ((y1 - YOFST) / TILEH, 0, MAPH-1); 
					yco <= BOUND ((y2 - YOFST) / TILEH, 0, MAPH-1); ++yco)
				{
					int i = xco + MAPW * yco;
					
					blit (sprites, canvas, TILEW * m[i], 0,
						XOFST + xco * TILEW, YOFST + yco * TILEH,
						TILEW, TILEH);
				}
			}
		}
		
		void Draw(BITMAP *canvas) 
		{
			if (screenDirty)
			{
				DrawBackground (canvas, 0, 0, canvas->w, canvas->h);
				screenDirty = false;
			}
			if (introtime > time)
			{
				if ((introtime - time) % 20 == 0)
				{
					textprintf_centre_ex (canvas, qfont, 
						320, 240, 
						makecol (255, 255, 255), makecol(0, 0, 0), 
						"READY?");
				}
				if ((introtime - time) % 20 == 10)
				{
					DrawBackground (canvas, 220, 240, 420, 248);
				}
			}			
			list<PcManObject>::iterator o;
			for (o = lobj.begin(); o != lobj.end(); ++o)
			{
				if (o->dirty)
				{
					DrawBackground (canvas, o->oldx + XOFST, o->oldy + YOFST, 
						o->oldx + TILEW + XOFST, o->oldy + TILEH + YOFST);
				}
			}
			for (o = lobj.begin(); o != lobj.end(); ++o)
			{
				int alt_sprite; // normal =1 or scared =2 ?
				int d = o->scared - time;
				if (d < 0) alt_sprite = 1;
					// start blinking in the last 100 ticks
				else if (d < 100) alt_sprite = (d % 16 < 8) ? 1 : 2;
				else alt_sprite = 2;
				blit (sprites, canvas,
					o->sprite * TILEW, alt_sprite * TILEH,
					XOFST + o->getx(), YOFST + o->gety(),
					TILEW, TILEH);
				o->dirty = false;
			}
			
			textprintf_centre_ex (canvas, qfont, 
				320, 464, 
				makecol (255, 255, 255), makecol(0, 0, 0), 
				"SCORE: %05i LIFE: %02i LEVEL: %02i", score, lives, level);
			
			// display debug messages
			//~ textprintf_centre_ex (canvas, qfont, 
				//~ 320, 472, 
				//~ makecol (255, 255, 255), makecol(0, 0, 0), 
				//~ "Left: %03i Level: %03i", left, obj[0].gety());
		}

		void HandlePacman ( PcManObject *o, int keys[8] )
		{
			int mp = (o->getmx()) + MAPW * (o->getmy());
			// if crossing a tile boundary
			if ((o->getx() & 15) == 0 &&
				(o->gety() & 15) == 0) 
			{
				// eat pills
				if (m[mp] == 0) { m[mp] = 3; score += 10; left--; }
				
				// eat magic pill
				if (m[mp] == 2) 
				{	m[mp] = 3; 
					score += 100; 
					// make all ghosts scared
					list<PcManObject>::iterator i;
					for (i = lobj.begin(); i != lobj.end(); ++i)
					{
						if (i->type == PcManObject::GHOST) 
							i->scared = time + POWERTIME;   
					}
					multiplier = 100; // reset score multiplier
				}
				
				// respond to key input
				if (keys[2] && m[mp-MAPW] != 1 && m[mp-MAPW] != 4) { o->dx = 0; o->dy = -1; }
				if (keys[3] && m[mp+MAPW] != 1 && m[mp+MAPW] != 4) { o->dx = 0; o->dy = 1; }
				if (keys[1] && m[mp+1] != 1 && m[mp+1] != 4) { o->dy = 0; o->dx = 1; }
				if (keys[0] && m[mp-1] != 1 && m[mp-1] != 4) { o->dy = 0; o->dx = -1; }
				
				// check if next tile is walkable
				{
					int dd;
					if (o->dx == 0) dd = 32 * o->dy; else dd = o->dx;
					if (m[mp + dd] == 1 || m[mp + dd] == 4) { o->dx = 0; o->dy = 0; }
				}
			}
		}
		
		void HandleGhost ( PcManObject *o )
		{
			int mp = (o->getmx()) + MAPW * (o->getmy());
			
			// if collission between ghost and player
			if (player->getx() < o->getx() + TILEW &&
				player->getx() + TILEW > o->getx() &&
				player->gety() < o->gety() + TILEH &&
				player->gety() + TILEH > o->gety() && 
				riptime < time)
			{
				// if power pill not enabled
				if (o->scared < time)
				{
					lives--;
					riptime = time + 50; // set a short delay
					// before resetting objects
				}
				else
				{
					o->setx(GHOST_STARTX * TILEW);
					o->sety(GHOST_STARTY * TILEH);
					score += multiplier;
					multiplier *= 2;
					o->scared = -1; // turn of scaredness
				}
			}
			
			// if crossing a tile boundary
			if ((o->getx() & 15) == 0 &&
				(o->gety() & 15) == 0) 
			{
				{
					int ndx[] = {1,0,-1,0};
					int ndy[] = {0,1,0,-1};
					int nmdd[] = {1,MAPW,-1,-MAPW};
					int nd;
					// random choice of 50%
					if (rand() % 1024 > 512)
						nd = (rand() % 1024) / 256;
					else
						nd = ((player->getx() > o->getx()) ^ (o->scared >= time)) ? 0 : 2;
					int j = 0;
					
					// while we've not tried all four,
					// and the map is not walkable in that dir,
					// or we're crossing a doorway in any dir other than right
					// or we're going backwards,
					while (j < 4 && (m[mp + nmdd[nd]] == 1) ||
						(ndx[nd] == -o->dx && ndy[nd] == -o->dy) ||
						(nd != 0 && m[mp + nmdd[nd]] == 4)
						) { nd = (++nd) & 3; j--; }
					o->dx = ndx[nd];
					o->dy = ndy[nd];
				}
			}
		}
		
		void MsgTimer ( int keys[8] )
		{
			if (left == 0)
			{
				level++;
				InitLevel();
			}							
			if (introtime > time)
			{
				// nothing, blink message
			}
			else if (riptime == time)
			{
				InitObjects();
				screenDirty = true;
			}
			else
			{
				list<PcManObject>::iterator o;
				for (o = lobj.begin(); o != lobj.end(); ++o)
				{
					// check if we cross a tile boundary
					switch (o->type)
					{
						case PcManObject::PACMAN: 
							HandlePacman (&*o, keys);
						break;
						case PcManObject::GHOST:
							HandleGhost (&*o);
						break;
						default:
							//we're not supposed to get here
							assert (false);
					}
					// move only if not in rip-time
					if (riptime < time && lives > 0) { 
						o->move (o->dx * 2, o->dy * 2);  // new position	
						if (o->gety() >= (TILEH * MAPH)) o->move (0, - TILEH * MAPH);
						if (o->gety() < 0) o->move (0, TILEH * MAPH);
					}
				}
			}
			time++;
		}
		
		int Logic(int keys[8], int mouse[4], float dT) {
			// Do all input processing and game logic here. Refer to the
			// comments in IGame.h for details on the format of the input
			// parameters and the requirements for the output.
			// ...
			while (timer > 0) {
				if (!paused) {
					MsgTimer(keys);
				}
				--timer;
			}
			if (lives > 0) return 0; else return 1;
		}

		int Score() {
			return score;
		}

		bool UseSystemMouse() {
			return false;
		}

		void Pause() {
			paused = true;
		}

		void Resume() {
			screenDirty = true;
			paused = false;
		}

		// Should return true if keys can be redefined (otherwise
		// global keys are used)
		bool RedefinesKeys() {
			return false;
		}

		// Should return names of redefinable keys or NULL if
		// key can't be redefined. i = [0,7]
		const char *KeyName(int i) {
			switch (i) {
				default: return 0;
			};
		}

		// Should return to a NULL terminated array of GameSettings structures.
		// For example you could define the array like this:
		//
		//	GameSettings my_settings[] = {
		//		{ "name1", "option 1", "first|second|third", 1 },
		//		{ "name2", "option 2", "first|second|third", 0 },
		//		{ "name3", "option 3", "first|second|third", 2 },
		//		{ 0, 0, 0, 0 }
		//	};
		//
		// and then return my_settings in this function.
		GameSettings* GetSettings() {
			return 0;
		}
};

extern "C" DLLEXPORT IGame* GetPlugin() {
	return new PcMan;
}

// for testing:
//~ const int PcMan::mapDefault [MAPW * MAPH] =
//~ {
	//~ 1,1,1,1,1,1,1,1,1,1,1,1,3,3,3,1,3,1,3,3,3,1,1,1,1,1,1,1,1,1,1,3,
	//~ 1,3,3,3,3,1,1,2,3,3,3,1,3,3,3,1,3,1,3,3,3,1,3,3,3,3,3,2,3,3,1,3,
	//~ 1,3,1,1,3,1,1,3,1,1,3,1,3,3,3,1,3,1,3,3,3,1,3,1,1,3,1,1,1,3,1,3,
	//~ 1,3,1,1,3,3,3,3,1,1,3,1,3,3,3,1,3,1,3,3,3,1,3,1,1,3,1,3,1,3,1,3,
	//~ 1,3,1,1,3,1,1,1,1,1,3,1,3,3,3,1,3,1,3,3,3,1,3,1,1,3,1,3,1,3,1,3,
	//~ 1,3,1,1,3,1,1,1,1,1,3,1,1,1,1,1,3,1,1,1,1,1,3,1,1,3,1,1,1,3,1,3,
	//~ 1,3,1,1,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,1,3,
	//~ 1,3,1,1,1,1,1,3,1,1,3,1,1,1,1,1,3,1,1,1,1,1,1,1,1,3,1,1,1,3,1,3,
	//~ 1,3,1,1,1,1,1,3,1,1,3,1,1,1,1,1,3,1,1,1,1,1,1,1,1,3,1,3,1,3,1,3,
	//~ 1,3,1,1,3,3,3,3,1,1,3,3,3,0,0,0,0,0,0,0,1,1,3,3,3,3,1,3,1,3,1,3,
	//~ 1,3,1,1,3,1,1,3,1,1,3,1,1,0,1,1,1,1,1,0,1,1,3,1,1,3,1,1,1,3,1,3,
	//~ 1,3,3,3,3,1,1,3,3,3,3,1,1,0,1,3,3,3,1,0,3,3,3,1,1,3,3,3,3,3,1,3,
	//~ 1,3,1,1,1,1,1,3,1,1,1,1,1,0,1,3,3,3,4,0,1,1,1,1,1,3,1,1,1,1,1,3
//~ };

const int PcMan::mapDefault [MAPW * MAPH] =
{
	1,1,1,1,1,1,1,1,1,1,1,1,3,3,3,1,0,1,3,3,3,1,1,1,1,1,1,1,1,1,1,3,
	1,0,0,0,0,1,1,2,0,0,0,1,3,3,3,1,0,1,3,3,3,1,0,0,0,0,0,2,0,0,1,3,
	1,0,1,1,0,1,1,0,1,1,0,1,3,3,3,1,0,1,3,3,3,1,0,1,1,0,1,1,1,0,1,3,
	1,0,1,1,0,0,0,0,1,1,0,1,3,3,3,1,0,1,3,3,3,1,0,1,1,0,1,3,1,0,1,3,
	1,0,1,1,0,1,1,1,1,1,0,1,3,3,3,1,0,1,3,3,3,1,0,1,1,0,1,3,1,0,1,3,
	1,0,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,1,0,1,1,1,0,1,3,
	1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,3,
	1,0,1,1,1,1,1,0,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,1,1,1,0,1,3,
	1,0,1,1,1,1,1,0,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,1,3,1,0,1,3,
	1,0,1,1,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,1,3,1,0,1,3,
	1,0,1,1,0,1,1,0,1,1,0,1,1,0,1,1,1,1,1,0,1,1,0,1,1,0,1,1,1,0,1,3,
	1,0,0,0,0,1,1,0,0,0,0,1,1,0,1,3,3,3,1,0,0,0,0,1,1,0,0,0,0,0,1,3,
	1,0,1,1,1,1,1,0,1,1,1,1,1,0,1,3,3,3,4,0,1,1,1,1,1,0,1,1,1,1,1,3
};

